declare type Cell = number;
declare type Matrix = any;
declare type Shape = number[];
declare type CellIndices = number[];
/**
 * Gets the matrix's shape.
 *
 * @param m
 */
declare function shape(m: Matrix): Shape;
/**
 * Validates that matrices are of the same shape.
 *
 * @param a
 * @param  b
 * @throws Error
 */
declare function validateSameShape(a: Matrix, b: Matrix): void;
/**
 * Generates the matrix of specific shape with specific values.
 *
 * @param mShape the shape of the matrix to generate
 * @param fill cell values of a generated matrix.
 */
declare function generate(mShape: Shape, fill: (recIndices: CellIndices) => Cell): Matrix;
/**
 * Generates the matrix of zeros of specified shape.
 *
 * @param mShape shape of the matrix
 */
declare function zeros(mShape: Shape): Matrix;
/**
 * @param a
 * @param b
 * @throws Error
 */
declare function dot(a: Matrix, b: Matrix): Matrix;
/**
 * Transposes the matrix.
 *
 * @param m
 * @throws Error
 */
declare function t(m: Matrix): Matrix;
/**
 * Traverses the matrix.
 *
 * @param m
 * @param visit
 */
declare function walk(m: Matrix, visit: (indices: CellIndices, c: Cell) => void): void;
/**
 * Gets the matrix cell value at specific index.
 *
 * @param m Matrix that contains the cell that needs to be updated
 * @param cellIndices Array of cell indices
 */
declare function getCellAtIndex(m: Matrix, cellIndices: CellIndices): Cell;
/**
 * Update the matrix cell at specific index.
 *
 * @param m Matrix that contains the cell that needs to be updated
 * @param cellIndices Array of cell indices
 * @param cellValue New cell value
 */
declare function updateCellAtIndex(m: Matrix, cellIndices: CellIndices, cellValue: Cell): void;
/**
 * Adds two matrices element-wise.
 *
 * @param a
 * @param b
 */
declare function add(a: Matrix, b: Matrix): Matrix;
/**
 * Multiplies two matrices element-wise.
 *
 * @param a
 * @param b
 */
declare function mul(a: Matrix, b: Matrix): Matrix;
/**
 * Subtract two matrices element-wise.
 *
 * @param a
 * @param b
 */
declare function sub(a: Matrix, b: Matrix): Matrix;

type Matrix$1_Matrix = Matrix;
declare const Matrix$1_shape: typeof shape;
declare const Matrix$1_validateSameShape: typeof validateSameShape;
declare const Matrix$1_generate: typeof generate;
declare const Matrix$1_zeros: typeof zeros;
declare const Matrix$1_dot: typeof dot;
declare const Matrix$1_t: typeof t;
declare const Matrix$1_walk: typeof walk;
declare const Matrix$1_getCellAtIndex: typeof getCellAtIndex;
declare const Matrix$1_updateCellAtIndex: typeof updateCellAtIndex;
declare const Matrix$1_add: typeof add;
declare const Matrix$1_mul: typeof mul;
declare const Matrix$1_sub: typeof sub;
declare namespace Matrix$1 {
  export {
    Matrix$1_Matrix as Matrix,
    Matrix$1_shape as shape,
    Matrix$1_validateSameShape as validateSameShape,
    Matrix$1_generate as generate,
    Matrix$1_zeros as zeros,
    Matrix$1_dot as dot,
    Matrix$1_t as t,
    Matrix$1_walk as walk,
    Matrix$1_getCellAtIndex as getCellAtIndex,
    Matrix$1_updateCellAtIndex as updateCellAtIndex,
    Matrix$1_add as add,
    Matrix$1_mul as mul,
    Matrix$1_sub as sub,
  };
}

/**
 * 凯撒密码加密
 *
 * @param str
 * @param shift
 * @param alphabet
 */
declare const caesarCipherEncrypt: (str: string, shift: number, alphabet?: string[]) => string;
/**
 * 凯撒密码解密
 *
 * @param str
 * @param shift
 * @param alphabet
 */
declare const caesarCipherDecrypt: (str: string, shift: number, alphabet?: string[]) => string;

/**
 * Encrypts the given message using Hill Cipher.
 *
 * @param message plaintext
 * @param keyString
 */
declare function hillCipherEncrypt(message: string, keyString: string): string;
declare function hillCipherDecrypt(): void;

declare class PolynomialHash {
    /**
     * Base number that is used to create the polynomial.
     */
    base: number;
    /**
     * Modulus number that keeps the hash from overflowing.
     */
    modulus: number;
    constructor({ base, modulus }?: {
        /**
         * Base number that is used to create the polynomial.
         */
        base?: number;
        /**
         * Modulus number that keeps the hash from overflowing.
         */
        modulus?: number;
    });
    /**
     * Function that creates hash representation of the word.
     *
     * Time complexity: O(word.length).
     *
     * @param word String that needs to be hashed.
     */
    hash(word: string): number;
    /**
     * Function that creates hash representation of the word
     * based on previous word (shifted by one character left) hash value.
     *
     * Recalculates the hash representation of a word so that it isn't
     * necessary to traverse the whole word again.
     *
     * Time complexity: O(1).
     *
     * @param prevHash
     * @param prevWord
     * @param newWord
     */
    roll(prevHash: number, prevWord: string, newWord: string): number;
    /**
     * Converts char to number.
     *
     * @param char
     */
    charToNumber(char: string): number;
}

declare class SimplePolynomialHash {
    /**
     * Base number that is used to create the polynomial.
     */
    base: number;
    /**
     * @param base Base number that is used to create the polynomial.
     */
    constructor(base?: number);
    /**
     * Function that creates hash representation of the word.
     *
     * Time complexity: O(word.length).
     *
     * @assumption: This version of the function  doesn't use modulo operator.
     * Thus it may produce number overflows by generating numbers that are
     * bigger than Number.MAX_SAFE_INTEGER. This function is mentioned here
     * for simplicity and LEARNING reasons.
     *
     * @param word String that needs to be hashed.
     */
    hash(word: string): number;
    /**
     * Function that creates hash representation of the word
     * based on previous word (shifted by one character left) hash value.
     *
     * Recalculates the hash representation of a word so that it isn't
     * necessary to traverse the whole word again.
     *
     * Time complexity: O(1).
     *
     * @assumption: This function doesn't use modulo operator and thus is not safe since
     * it may deal with numbers that are bigger than Number.MAX_SAFE_INTEGER. This
     * function is mentioned here for simplicity and LEARNING reasons.
     *
     * @param prevHash
     * @param prevWord
     * @param newWord
     */
    roll(prevHash: number, prevWord: string, newWord: string): number;
}

/**
 * Encodes the message using Rail Fence Cipher.
 *
 * @param string The string to be encoded
 * @param railCount The number of rails in a fence
 * @returns Encoded string
 */
declare function encodeRailFenceCipher(string: string, railCount: number): string;
/**
 * Decodes the message using Rail Fence Cipher.
 *
 * @param string Encoded string
 * @param railCount The number of rows in a fence
 * @returns Decoded string.
 */
declare function decodeRailFenceCipher(string: string, railCount: number): string;

/**
 * 链表结点
 */
declare class LinkedListNode<T> {
    /**
     * 值
     */
    value: T;
    /**
     * 下一个结点
     */
    next: LinkedListNode<T>;
    constructor(value: T, next?: any);
    toString(callback?: (value: T) => string): string;
}

declare type CompareFunction<T> = (a: T, b: T) => number;
/**
 * 比较器
 */
declare class Comparator<T> {
    private compare;
    /**
     * Constructor.
     *
     * @param compareFunction It may be custom compare function that, let's say may compare custom objects together.
     */
    constructor(compareFunction?: CompareFunction<T>);
    /**
     * Default comparison function. It just assumes that "a" and "b" are strings or numbers.
     * @param a
     * @param b
     * @returns
     */
    static defaultCompareFunction(a: string | number, b: string | number): 0 | 1 | -1;
    /**
     * Checks if two variables are equal.
     *
     * @param a
     * @param b
     * @return
     */
    equal(a: T, b: T): boolean;
    /**
     * Checks if variable "a" is less than "b".
     *
     * @param a
     * @param  b
     * @return
     */
    lessThan(a: T, b: T): boolean;
    /**
     * Checks if variable "a" is greater than "b".
     *
     * @param a
     * @param  b
     * @return
     */
    greaterThan(a: T, b: T): boolean;
    /**
     * Checks if variable "a" is less than or equal to "b".
     *
     * @param a
     * @param  b
     * @return
     */
    lessThanOrEqual(a: T, b: T): boolean;
    /**
     * Checks if variable "a" is greater than or equal to "b".
     *
     * @param a
     * @param b
     * @return
     */
    greaterThanOrEqual(a: T, b: T): boolean;
    /**
     * Reverses the comparison order.
     */
    reverse(): void;
}

/**
 * 链表
 */
declare class LinkedList<T> {
    /**
     * 表头
     */
    head: LinkedListNode<T>;
    /**
     * 表尾
     */
    tail: LinkedListNode<T>;
    /**
     * 比较器
     */
    private compare;
    /**
     * 构建链表
     *
     * @param comparatorFunction 比较函数
     */
    constructor(comparatorFunction?: CompareFunction<T>);
    /**
     * 添加新结点到表头
     *
     * @param value 新结点值
     * @return 新增的结点
     */
    prepend(value: T): this;
    /**
     * 添加新结点到表尾
     *
     * @param value 新结点值
     * @return 新增的结点
     */
    append(value: T): this;
    /**
     * 插入新结点
     *
     * @param value 新结点值
     * @param rawIndex 插入索引
     * @return 自身
     */
    insert(value: T, rawIndex: number): this;
    /**
     * 删除链表中第一个与指定值相等的结点
     *
     * @param value 结点值
     * @return 被删除的结点
     */
    delete(value: T): LinkedListNode<T>;
    /**
     * 查找结点
     *
     * @param {Object} findParams
     * @param {*} findParams.value
     * @param {function} [findParams.callback]
     *
     * @return 查找到的结点
     */
    find({ value, callback }: {
        /**
         * 结点值
         */
        value?: T;
        /**
         * 比较函数
         */
        callback?: (value: T) => boolean;
    }): LinkedListNode<T>;
    /**
     * 删除表尾
     *
     * @return 被删除的结点
     */
    deleteTail(): LinkedListNode<T>;
    /**
     * 删除表头
     *
     * @return 被删除结点
     */
    deleteHead(): LinkedListNode<T>;
    /**
     * 从数组中初始化链表
     *
     * @param values - Array of values that need to be converted to linked list.
     * @return 自身
     */
    fromArray(values: T[]): this;
    /**
     * 转换为数组
     */
    toArray(): LinkedListNode<T>[];
    /**
     * 转换为字符串
     *
     * @param callback
     */
    toString(callback?: (value: T) => string): string;
    /**
     * Reverse a linked list.
     *
     * @returns 自身
     */
    reverse(): this;
}

/**
 * 图顶点
 */
declare class GraphVertex<T> {
    /**
     * 值
     */
    value: T;
    /**
     * 边列表
     */
    edges: LinkedList<GraphEdge<T>>;
    /**
     * 构建图顶点
     *
     * @param value 值
     */
    constructor(value?: T);
    /**
     * 新增边
     *
     * @param edge 新增的边
     * @returns 自身
     */
    addEdge(edge: GraphEdge<T>): this;
    /**
     * 删除边
     *
     * @param edge 边
     */
    deleteEdge(edge: GraphEdge<T>): void;
    /**
     * 获取相邻顶点列表
     */
    getNeighbors(): GraphVertex<T>[];
    /**
     * 获取边列表
     */
    getEdges(): GraphEdge<T>[];
    /**
     * 获取边的数量
     */
    getDegree(): number;
    /**
     * 是否存在指定边
     *
     * @param requiredEdge 指定边
     * @returns 是否存在指定边
     */
    hasEdge(requiredEdge: GraphEdge<T>): boolean;
    /**
     * 是否有相邻顶点
     *
     * @param vertex 顶点
     * @returns 是否有相邻顶点
     */
    hasNeighbor(vertex: GraphVertex<T>): boolean;
    /**
     * 查找边
     *
     * @param vertex 顶点
     * @returns 查找到的边
     */
    findEdge(vertex: GraphVertex<T>): GraphEdge<T>;
    /**
     * 获取键值
     */
    getKey(): string;
    /**
     * 删除所有边
     */
    deleteAllEdges(): this;
    /**
     * 转换为字符串
     *
     * @param callback 转换为字符串函数
     */
    toString(callback?: (value: T) => string): string;
}

/**
 * 图边
 */
declare class GraphEdge<T> {
    /**
     * 起始顶点
     */
    startVertex: GraphVertex<T>;
    /**
     * 结束顶点
     */
    endVertex: GraphVertex<T>;
    /**
     * 权重
     */
    weight: number;
    /**
     * 构建图边
     * @param startVertex 起始顶点
     * @param endVertex 结束顶点
     * @param weight 权重
     */
    constructor(startVertex: GraphVertex<T>, endVertex: GraphVertex<T>, weight?: number);
    /**
     * 获取键值
     */
    getKey(): string;
    /**
     * 反转
     */
    reverse(): this;
    /**
     * 转换为字符串
     */
    toString(): string;
}

/**
 * 图
 */
declare class Graph<T> {
    /**
     * 顶点列表
     */
    vertices: {
        [key: string]: GraphVertex<T>;
    };
    /**
     * 边列表
     */
    edges: {
        [key: string]: GraphEdge<T>;
    };
    /**
     * 是否有向
     */
    isDirected: boolean;
    /**
     * 构建图
     *
     * @param isDirected 是否有向
     */
    constructor(isDirected?: boolean);
    /**
     * 新增顶点
     *
     * @param newVertex 新顶点
     */
    addVertex(newVertex: GraphVertex<T>): this;
    /**
     * 获取顶点
     *
     * @param vertexKey 顶点键值
     */
    getVertexByKey(vertexKey: string): GraphVertex<T>;
    /**
     * 获取相邻点
     *
     * @param vertex 顶点
     */
    getNeighbors(vertex: GraphVertex<T>): GraphVertex<T>[];
    /**
     * 获取所有顶点
     */
    getAllVertices(): GraphVertex<T>[];
    /**
     * 获取所有边
     */
    getAllEdges(): GraphEdge<T>[];
    /**
     * 新增边
     *
     * @param edge 边
     */
    addEdge(edge: GraphEdge<T>): this;
    /**
     * 删除边
     *
     * @param edge 边
     */
    deleteEdge(edge: GraphEdge<T>): void;
    /**
     * 查找边
     *
     * @param ertex 起始顶点
     * @param endVertex 结束顶点
     * @return 查找到的边
     */
    findEdge(startVertex: GraphVertex<T>, endVertex: GraphVertex<T>): GraphEdge<T>;
    /**
     * 获取权重
     */
    getWeight(): number;
    /**
     * Reverse all the edges in directed graph.
     */
    reverse(): this;
    /**
     * 获取所有顶点索引
     */
    getVerticesIndices(): {
        [key: string]: number;
    };
    /**
     * 获取邻接矩阵
     */
    getAdjacencyMatrix(): number[][];
    /**
     * 转换为字符串
     */
    toString(): string;
}

/**
 * Tarjan's algorithm for finding articulation points in graph.
 *
 * @param graph
 */
declare function articulationPoints<T>(graph: Graph<T>): {
    [key: string]: GraphVertex<T>;
};

/**
 * @param graph
 * @param startVertex
 */
declare function bellmanFord<T>(graph: Graph<T>, startVertex: GraphVertex<T>): {
    distances: {
        [key: string]: number;
    };
    previousVertices: {
        [key: string]: GraphVertex<T>;
    };
};

interface Callbacks$2<T> {
    /**
     * Determines whether DFS should traverse from the vertex to its neighbor
     * (along the edge). By default prohibits visiting the same vertex again.
     */
    allowTraversal?: (vertices: {
        currentVertex: GraphVertex<T>;
        previousVertex: GraphVertex<T>;
        nextVertex: GraphVertex<T>;
    }) => boolean;
    /**
     * Called when BFS enters the vertex.
     */
    enterVertex?: (vertices: {
        currentVertex: GraphVertex<T>;
        previousVertex: GraphVertex<T>;
    }) => any;
    /**
     * Called when BFS leaves the vertex.
     */
    leaveVertex?: (vertices: {
        currentVertex: GraphVertex<T>;
        previousVertex: GraphVertex<T>;
    }) => void;
}
/**
 * @param graph
 * @param startVertex
 * @param originalCallbacks
 */
declare function breadthFirstSearch<T>(graph: Graph<T>, startVertex: GraphVertex<T>, originalCallbacks?: Callbacks$2<T>): void;

/**
 * @param graph
 */
declare function graphBridges<T>(graph: Graph<T>): {
    [key: string]: GraphEdge<T>;
};

interface Callbacks$1<T> {
    /**
     * Determines whether DFS should traverse from the vertex to its neighbor
     * (along the edge). By default prohibits visiting the same vertex again.
     */
    allowTraversal?: (vertices: any) => boolean;
    /**
     * Called when DFS enters the vertex.
     */
    enterVertex?: (vertices: any) => any;
    /**
     * Called when DFS leaves the vertex.
     */
    leaveVertex?: (vertices: {
        currentVertex: GraphVertex<T>;
        previousVertex: GraphVertex<T>;
    }) => any;
}
/**
 * @param graph
 * @param startVertex
 * @param callbacks
 */
declare function depthFirstSearch<T>(graph: Graph<T>, startVertex: GraphVertex<T>, callbacks?: Callbacks$1<T>): void;

/**
 * Detect cycle in directed graph using Depth First Search.
 *
 * @param graph
 */
declare function detectDirectedCycle<T>(graph: Graph<T>): {
    [key: string]: GraphVertex<T>;
};

/**
 * Detect cycle in undirected graph using Depth First Search.
 *
 * @param graph
 */
declare function detectUndirectedCycle<T>(graph: Graph<T>): {
    [key: string]: GraphVertex<T>;
};

/**
 * Detect cycle in undirected graph using disjoint sets.
 *
 * @param graph
 */
declare function detectUndirectedCycleUsingDisjointSet<T>(graph: Graph<T>): boolean;

interface ShortestPaths<T> {
    /**
     * shortest distances to all vertices
     */
    distances: {
        [key: string]: number;
    };
    /**
     * shortest paths to all vertices.
     */
    previousVertices: {
        [key: string]: GraphVertex<T>;
    };
}
/**
 * Implementation of Dijkstra algorithm of finding the shortest paths to graph nodes.
 * @param graph graph we're going to traverse.
 * @param startVertex traversal start vertex.
 */
declare function dijkstra<T>(graph: Graph<T>, startVertex: GraphVertex<T>): ShortestPaths<T>;

/**
 * Fleury's algorithm of finding Eulerian Path (visit all graph edges exactly once).
 *
 * @param graph
 */
declare function eulerianPath<T>(graph: Graph<T>): GraphVertex<T>[];

/**
 * @param graph
 */
declare function kruskal<T>(graph: Graph<T>): Graph<T>;

/**
 * @param graph
 */
declare function prim<T>(graph: Graph<T>): Graph<T>;

/**
 * Kosaraju's algorithm.
 *
 * @param graph
 */
declare function stronglyConnectedComponents<T>(graph: Graph<T>): GraphVertex<T>[][];

/**
 * @param graph
 */
declare function topologicalSort<T>(graph: Graph<T>): GraphVertex<T>[];

/**
 * BRUTE FORCE approach to solve Traveling Salesman Problem.
 *
 * @param graph
 */
declare function bfTravellingSalesman<T>(graph: Graph<T>): GraphVertex<T>[];

/**
 * Type that describes the image size (width and height)
 */
declare type ImageSize = {
    /**
     * image width.
     */
    w: number;
    /**
     * image height.
     */
    h: number;
};
declare type ResizeImageWidthArgs = {
    /**
     * image data we want to resize.
     */
    img: ImageData;
    /**
     * final image width we want the image to shrink to.
     */
    toWidth: number;
};
declare type ResizeImageWidthResult = {
    /**
     * resized image data.
     */
    img: ImageData;
    /**
     * resized image size.
     */
    size: ImageSize;
};
/**
 * Performs the content-aware image width resizing using the seam carving method.
 *
 * @param args
 */
declare function resizeImageWidth({ img, toWidth }: ResizeImageWidthArgs): ResizeImageWidthResult;

declare type PixelColor = ArrayLike<number> | Uint8ClampedArray;
declare type PixelCoordinate = {
    /**
     * horizontal coordinate.
     */
    x: number;
    /**
     * vertical coordinate.
     */
    y: number;
};
/**
 * Helper function that returns the color of the pixel.
 * @param img
 * @param coordinate
 */
declare const getPixel: (img: ImageData, { x, y }: PixelCoordinate) => Uint8ClampedArray;
/**
 * Helper function that sets the color of the pixel.
 *
 * @param img
 * @param coordinate
 * @param color
 */
declare const setPixel: (img: ImageData, { x, y }: PixelCoordinate, color: PixelColor) => void;

/**
 * Traversal callback function.
 */
declare type traversalCallback<T> = (nodeValue: T) => void;
/**
 * @param linkedList
 * @param callback
 */
declare function reverseTraversal<T>(linkedList: LinkedList<T>, callback: traversalCallback<T>): void;

/**
 * @param linkedList
 * @param callback
 */
declare function traversal<T>(linkedList: LinkedList<T>, callback: (nodeValue: T) => void): void;

/**
 * Sequence of 0s and 1s.
 */
declare type Bits = number[];
/**
 *  Converts the 16-bit binary representation of the floating point number to decimal float number.
 *
 * @param bits sequence of bits that represents the floating point number.
 * @return floating point number decoded from its binary representation.
 */
declare function bitsToFloat16(bits: Bits): number;
/**
 * Converts the 32-bit binary representation of the floating point number to decimal float number.
 *
 * @param bits sequence of bits that represents the floating point number.
 * @return floating point number decoded from its binary representation.
 */
declare function bitsToFloat32(bits: Bits): number;
/**
 * Converts the 64-bit binary representation of the floating point number to decimal float number.
 *
 * @param bits sequence of bits that represents the floating point number.
 * @return floating point number decoded from its binary representation.
 */
declare function bitsToFloat64(bits: Bits): number;

/**
 * Converts the float number into its IEEE 754 64-bits binary representation.
 *
 * @param floatNumber float number in decimal format.
 * @return 64 bits binary string representation of the float number.
 */
declare function floatAs64BinaryString(floatNumber: number): string;
/**
 * Converts the float number into its IEEE 754 32-bits binary representation.
 *
 * @param floatNumber float number in decimal format.
 * @return 32 bits binary string representation of the float number.
 */
declare function floatAs32BinaryString(floatNumber: number): string;

/**
 * Return the number of bits used in the binary representation of the number.
 *
 * @param number
 */
declare function bitLength(number: number): number;

/**
 * Counts the number of bits that need to be change in order
 * to convert numberA to numberB.
 *
 * @param numberA
 * @param numberB
 */
declare function bitsDiff(numberA: number, numberB: number): number;

/**
 * @param number
 * @param bitPosition zero based.
 */
declare function clearBit(number: number, bitPosition: number): number;

/**
 * @param originalNumber
 */
declare function countSetBits(originalNumber: number): number;

/**
 * @param number
 */
declare function divideByTwo(number: number): number;

/**
 * Add two numbers using only binary operators.
 *
 * This is an implementation of full adders logic circuit.
 * https://en.wikipedia.org/wiki/Adder_(electronics)
 * Inspired by: https://www.youtube.com/watch?v=wvJc9CZcvBc
 *
 * Table(1)
 *  INPUT  | OUT
 *  C Ai Bi | C Si | Row
 * -------- | -----| ---
 *  0  0  0 | 0  0 | 1
 *  0  0  1 | 0  1 | 2
 *  0  1  0 | 0  1 | 3
 *  0  1  1 | 1  0 | 4
 * -------- | ---- | --
 *  1  0  0 | 0  1 | 5
 *  1  0  1 | 1  0 | 6
 *  1  1  0 | 1  0 | 7
 *  1  1  1 | 1  1 | 8
 * ---------------------
 *
 * Legend:
 * INPUT C = Carry in, from the previous less-significant stage
 * INPUT Ai = ith bit of Number A
 * INPUT Bi = ith bit of Number B
 * OUT C = Carry out to the next most-significant stage
 * OUT Si = Bit Sum, ith least significant bit of the result
 *
 *
 * @param a
 * @param b
 */
declare function fullAdder(a: number, b: number): number;

/**
 * @param number
 * @param bitPosition zero based.
 */
declare function getBit(number: number, bitPosition: number): number;

/**
 * @param number
 */
declare function isEven(number: number): boolean;

/**
 * @param number 32-bit integer.
 */
declare function isPositive(number: number): boolean;

/**
 * @param number
 */
declare function isPowerOfTwo(number: number): boolean;

/**
 * Multiply two signed numbers using bitwise operations.
 *
 * If a is zero or b is zero or if both a and b are zeros:
 * multiply(a, b) = 0
 *
 * If b is even:
 * multiply(a, b) = multiply(2a, b/2)
 *
 * If b is odd and b is positive:
 * multiply(a, b) = multiply(2a, (b-1)/2) + a
 *
 * If b is odd and b is negative:
 * multiply(a, b) = multiply(2a, (b+1)/2) - a
 *
 * Time complexity: O(log b)
 *
 * @param a
 * @param b
 */
declare function multiply(a: number, b: number): number;

/**
 * @param number
 */
declare function multiplyByTwo(number: number): number;

/**
 * Multiply to unsigned numbers using bitwise operator.
 *
 * The main idea of bitwise multiplication is that every number may be split
 * to the sum of powers of two:
 *
 * I.e. 19 = 2^4 + 2^1 + 2^0
 *
 * Then multiplying number x by 19 is equivalent of:
 *
 * x * 19 = x * 2^4 + x * 2^1 + x * 2^0
 *
 * Now we need to remember that (x * 2^4) is equivalent of shifting x left by 4 bits (x << 4).
 *
 * @param number1
 * @param number2
 */
declare function multiplyUnsigned(number1: number, number2: number): number;

/**
 * @param number
 * @param bitPosition zero based.
 */
declare function setBit(number: number, bitPosition: number): number;

/**
 * Switch the sign of the number using "Twos Complement" approach.
 * @param number
 */
declare function switchSign(number: number): number;

/**
 * 复数 (数学)
 */
declare class ComplexNumber {
    re: number;
    im: number;
    /**
     * z = re + im * i
     * z = radius * e^(i * phase)
     *
     */
    constructor({ re, im }?: {
        re?: number;
        im?: number;
    });
    /**
     * @param addend
     */
    add(addend: ComplexNumber | number): ComplexNumber;
    /**
     * @param subtrahend
     */
    subtract(subtrahend: ComplexNumber | number): ComplexNumber;
    /**
     * @param multiplicand
     */
    multiply(multiplicand: ComplexNumber | number): ComplexNumber;
    /**
     * @param divider
     */
    divide(divider: ComplexNumber | number): ComplexNumber;
    /**
     * @param number
     */
    conjugate(number: ComplexNumber | number): ComplexNumber;
    getRadius(): number;
    /**
     * @param inRadians
     */
    getPhase(inRadians?: boolean): number;
    /**
     * @param inRadians
     */
    getPolarForm(inRadians?: boolean): {
        radius: number;
        phase: number;
    };
    /**
     * Convert real numbers to complex number.
     * In case if complex number is provided then lefts it as is.
     *
     * @param number
     */
    toComplexNumber(number: ComplexNumber | number): ComplexNumber;
}

/**
 * Recursive version of Euclidean Algorithm of finding greatest common divisor (GCD).
 * @param originalA
 * @param originalB
 */
declare function euclideanAlgorithm(originalA: number, originalB: number): any;

/**
 * Iterative version of Euclidean Algorithm of finding greatest common divisor (GCD).
 * @param originalA
 * @param originalB
 */
declare function euclideanAlgorithmIterative(originalA: number, originalB: number): number;

/**
 * @typedef {import('../matrix/Matrix.js').Matrix} Matrix
 */

/**
 * Calculates the euclidean distance between 2 matrices.
 *
 * @param a
 * @param b
 *
 * @trows Error
 */
declare function euclideanDistance(a: Matrix, b: Matrix): number;

/**
 * 阶乘
 * N!
 *
 * @param number 阶乘数
 * @returns 值
 */
declare function factorial(number: number): number;

/**
 * 阶乘（递归）
 *
 * @param number 阶乘数
 * @return 值
 */
declare function factorialRecursive(number: number): any;

/**
 * Fast Powering Algorithm.
 * Recursive implementation to compute power.
 *
 * Complexity: log(n)
 *
 * @param base Number that will be raised to the power.
 * @param power The power that number will be raised to.
 */
declare function fastPowering(base: number, power: number): number;

/**
 * Return a fibonacci sequence as an array.
 *
 * @param n
 */
declare function fibonacci(n: number): number[];

/**
 * Calculate fibonacci number at specific position using Dynamic Programming approach.
 *
 * @param n
 */
declare function fibonacciNth(n: number): number;

/**
 * Calculate fibonacci number at specific position using closed form function (Binet's formula).
 * @see: https://en.wikipedia.org/wiki/Fibonacci_number#Closed-form_expression
 *
 * @param position Position number of fibonacci sequence (must be number from 1 to 75).
 */
declare function fibonacciNthClosedForm(position: number): number;

/**
 * Discrete Fourier Transform (DFT): time to frequencies.
 *
 * Time complexity: O(N^2)
 *
 * @param inputAmplitudes Input signal amplitudes over time (complex
 * numbers with real parts only).
 * @param zeroThreshold Threshold that is used to convert real and imaginary numbers
 * to zero in case if they are smaller then this.
 *
 * @return Array of complex number. Each of the number represents the frequency
 * or signal. All signals together will form input signal over discrete time periods. Each signal's
 * complex number has radius (amplitude) and phase (angle) in polar form that describes the signal.
 *
 * @see https://gist.github.com/anonymous/129d477ddb1c8025c9ac
 * @see https://betterexplained.com/articles/an-interactive-guide-to-the-fourier-transform/
 */
declare function discreteFourierTransform(inputAmplitudes: number[], zeroThreshold?: number): ComplexNumber[];

/**
 * Returns the radix-2 fast fourier transform of the given array.
 * Optionally computes the radix-2 inverse fast fourier transform.
 *
 * @param inputData
 * @param inverse
 */
declare function fastFourierTransform(inputData: ComplexNumber[], inverse?: boolean): ComplexNumber[];

/**
 * Inverse Discrete Fourier Transform (IDFT): frequencies to time.
 *
 * Time complexity: O(N^2)
 *
 * @param frequencies Frequencies summands of the final signal.
 * @param zeroThreshold Threshold that is used to convert real and imaginary numbers
 * to zero in case if they are smaller then this.
 *
 * @return Discrete amplitudes distributed in time.
 */
declare function inverseDiscreteFourierTransform(frequencies: ComplexNumber[], zeroThreshold?: number): number[];

/**
 * Returns the evaluation of a polynomial function at a certain point.
 * Uses straightforward approach with powers.
 *
 * @param coefficients i.e. [4, 3, 2] for (4 * x^2 + 3 * x + 2)
 * @param xVal
 */
declare function classicPolynome(coefficients: number[], xVal: number): number;

/**
 * Returns the evaluation of a polynomial function at a certain point.
 * Uses Horner's rule.
 *
 * @param coefficients i.e. [4, 3, 2] for (4 * x^2 + 3 * x + 2)
 * @param xVal
 */
declare function hornerMethod(coefficients: number[], xVal: number): number;

/**
 * @param number
 */
declare function integerPartition(number: number): number;

/**
 * 是否为2的幂 2^n
 *
 * @param number 被检查的值
 */
declare function isPowerOfTwo1(number: number): boolean;

/**
 * 是否为2的幂 2^n
 *
 * @param number 被检查的值
 */
declare function isPowerOfTwoBitwise(number: number): boolean;

/**
 * @param a
 * @param b
 */
declare function leastCommonMultiple(a: number, b: number): number;

/**
 * Calculate the π value using Liu Hui's π algorithm
 *
 * @param splitCount number of times we're going to split 6-gon.
 *  On each split we will receive 12-gon, 24-gon and so on.
 */
declare function liuHui(splitCount?: number): number;

/**
 * @param lineNumber - zero based.
 */
declare function pascalTriangle(lineNumber: number): number[];

/**
 * @param lineNumber zero based.
 */
declare function pascalTriangleRecursive(lineNumber: number): number[];

/**
 * @param number
 */
declare function trialDivision(number: number): boolean;

/**
 * Finds prime factors of a number.
 *
 * @param n the number that is going to be split into prime factors.
 * @returns array of prime factors.
 */
declare function primeFactors(n: number): number[];
/**
 * Hardy-Ramanujan approximation of prime factors count.
 *
 * @param n
 * @returns approximate number of prime factors.
 */
declare function hardyRamanujan(n: number): number;

/**
 * @param degree
 */
declare function degreeToRadian(degree: number): number;

/**
 * @param radian
 */
declare function radianToDegree(radian: number): number;

/**
 * @param maxNumber
 */
declare function sieveOfEratosthenes(maxNumber: number): number[];

/**
 * Calculates the square root of the number with given tolerance (precision)
 * by using Newton's method.
 *
 * @param number the number we want to find a square root for.
 * @param tolerance how many precise numbers after the floating point we want to get.
 */
declare function squareRoot(number: number, tolerance?: number): number;

/**
 * Classifies the point in space based on k-Means algorithm.
 *
 * @param data array of dataSet points, i.e. [[0, 1], [3, 4], [5, 7]]
 * @param k number of clusters
 * @return the class of the point
 */
declare function KMeans(data: number[][], k?: number): number[];

/**
 * Classifies the point in space based on k-nearest neighbors algorithm.
 *
 * @param dataSet array of data points, i.e. [[0, 1], [3, 4], [5, 7]]
 * @param labels array of classes (labels), i.e. [1, 1, 2]
 * @param toClassify the point in space that needs to be classified, i.e. [5, 4]
 * @param k number of nearest neighbors which will be taken into account (preferably odd)
 * @return the class of the point
 */
declare function kNN(dataSet: number[][], labels: number[], toClassify: number[], k?: number): number;

/**
 * Binary search implementation.
 *
 * @param sortedArray
 * @param seekElement
 * @param comparatorCallback
 */
declare function binarySearch<T>(sortedArray: T[], seekElement: T, comparatorCallback?: (a: T, b: T) => number): number;

/**
 * Interpolation search implementation.
 *
 * @param sortedArray sorted array with uniformly distributed values
 * @param seekElement
 */
declare function interpolationSearch(sortedArray: number[], seekElement: number): number;

/**
 * Jump (block) search implementation.
 *
 * @param sortedArray
 * @param seekElement
 * @param comparatorCallback
 */
declare function jumpSearch<T>(sortedArray: T[], seekElement: T, comparatorCallback?: (a: T, b: T) => number): number;

/**
 * Linear search implementation.
 *
 * @param array
 * @param seekElement
 * @param comparatorCallback
 */
declare function linearSearch<T>(array: T[], seekElement: T, comparatorCallback?: (a: T, b: T) => number): number[];

/**
 * Generates Cartesian Product of two sets.
 *
 * @param setA
 * @param setB
 */
declare function cartesianProduct<T>(setA: T[], setB: T[]): T[][];

/**
 * Backtracking algorithm of finding all possible combination for specific sum.
 *
 * @param candidates
 * @param target
 */
declare function combinationSum(candidates: number[], target: number): number[][];

/**
 * @param comboOptions
 * @param comboLength
 */
declare function combineWithoutRepetitions<T>(comboOptions: T[], comboLength: number): T[][];

/**
 * @param comboOptions
 * @param comboLength
 */
declare function combineWithRepetitions<T>(comboOptions: T[], comboLength: number): T[][];

/**
 * @param originalArray
 */
declare function fisherYates<T>(originalArray: T[]): T[];

declare class KnapsackItem {
    /**
     * value of the item.
     */
    value: number;
    /**
     * weight of the item.
     */
    weight: number;
    /**
     * how many items are available to be added.
     */
    itemsInStock: number;
    /**
     * Actual number of items that is going to be added to knapsack.
     */
    quantity: number;
    /**
     */
    constructor({ value, weight, itemsInStock }: {
        /**
         * value of the item.
         */
        value: number;
        /**
         * weight of the item.
         */
        weight: number;
        /**
         * how many items are available to be added.
         */
        itemsInStock?: number;
    });
    get totalValue(): number;
    get totalWeight(): number;
    get valuePerWeightRatio(): number;
    toString(): string;
}

declare class Knapsack {
    selectedItems: KnapsackItem[];
    weightLimit: number;
    possibleItems: KnapsackItem[];
    /**
     * @param {KnapsackItem[]} possibleItems
     * @param {number} weightLimit
     */
    constructor(possibleItems: KnapsackItem[], weightLimit: number);
    sortPossibleItemsByWeight(): void;
    sortPossibleItemsByValue(): void;
    sortPossibleItemsByValuePerWeightRatio(): void;
    solveZeroOneKnapsackProblem(): void;
    solveUnboundedKnapsackProblem(): void;
    get totalValue(): number;
    get totalWeight(): number;
}

/**
 * @param set1
 * @param set2
 */
declare function longestCommonSubsequence(set1: string[], set2: string[]): string[];

/**
 * Dynamic programming approach to find longest increasing subsequence.
 * Complexity: O(n * n)
 *
 * @param sequence
 */
declare function dpLongestIncreasingSubsequence(sequence: number[]): number;

/**
 * Brute Force solution.
 * Complexity: O(n^2)
 *
 * @param inputArray
 */
declare function bfMaximumSubarray(inputArray: number[]): number[];

/**
 * Divide and Conquer solution.
 * Complexity: O(n^2) in case if no memoization applied
 *
 * @param inputArray
 */
declare function dcMaximumSubarraySum(inputArray: number[]): number;

/**
 * Dynamic Programming solution.
 * Complexity: O(n)
 *
 * @param inputArray
 */
declare function dpMaximumSubarray(inputArray: number[]): number[];

/**
 * @param permutationOptions
 */
declare function permutateWithoutRepetitions<T>(permutationOptions: T[]): T[][];

/**
 * @param permutationOptions
 * @param permutationLength
 */
declare function permutateWithRepetitions<T>(permutationOptions: T[], permutationLength?: number): T[][];

/**
 * Find power-set of a set using BACKTRACKING approach.
 *
 * @param originalSet
 */
declare function btPowerSet<T>(originalSet: T[]): T[][];

/**
 * Find power-set of a set using BITWISE approach.
 *
 * @param originalSet
 */
declare function bwPowerSet<T>(originalSet: T[]): T[][];

/**
 * @param set1
 * @param set2
 */
declare function shortestCommonSupersequence(set1: string[], set2: string[]): string[];

interface SorterCallbacks<T> {
    /**
     * If provided then all elements comparisons
     * will be done through this callback.
     */
    compareCallback?: CompareFunction<T>;
    /**
     * If provided it will be called each time the sorting
     * function is visiting the next element.
     */
    visitingCallback?: (v?: T) => void;
}
declare class Sort<T> {
    callbacks: SorterCallbacks<T>;
    comparator: Comparator<T>;
    constructor(originalCallbacks?: SorterCallbacks<T>);
    /**
     * @param originalCallbacks
     */
    static initSortingCallbacks<T>(originalCallbacks: SorterCallbacks<T>): SorterCallbacks<T>;
    sort(_array?: T[]): T[];
}

declare class BubbleSort<T> extends Sort<T> {
    sort(originalArray?: T[]): T[];
}

/**
 * 计数排序（Counting sort）
 */
declare class CountingSort extends Sort<number> {
    /**
     * @param originalArray
     * @param smallestElement
     * @param biggestElement
     */
    sort(originalArray: number[], smallestElement?: number, biggestElement?: number): number[];
}

declare class InsertionSort<T> extends Sort<T> {
    sort(originalArray: T[]): T[];
}

declare class MergeSort<T> extends Sort<T> {
    sort(originalArray: T[]): T[];
    mergeSortedArrays(leftArray: T[], rightArray: T[]): T[];
}

declare class QuickSort<T> extends Sort<T> {
    /**
     * @param originalArray
     */
    sort(originalArray: T[]): T[];
}

declare class QuickSortInPlace<T> extends Sort<T> {
    /** Sorting in place avoids unnecessary use of additional memory, but modifies input array.
     *
     * This process is difficult to describe, but much clearer with a visualization:
     * @see: http://www.algomation.com/algorithm/quick-sort-visualization
     *
     * @param originalArray Not sorted array.
     * @param inputLowIndex
     * @param inputHighIndex
     * @param recursiveCall
     * @return Sorted array.
     */
    sort(originalArray: T[], inputLowIndex?: number, inputHighIndex?: number, recursiveCall?: boolean): T[];
}

/**
 * 基数排序（Radix sort）
 */
declare class RadixSort<T> extends Sort<T> {
    /**
     * @param originalArray
     */
    sort(originalArray: T[]): T[];
    /**
     * @param array
     * @param index
     */
    placeElementsInNumberBuckets(array: T[], index: number): T[][];
    /**
     * @param array
     * @param index
     * @param numPasses
     */
    placeElementsInCharacterBuckets(array: T[], index: number, numPasses: number): T[][];
    /**
     * @param element
     * @param index
     * @param numPasses
     */
    getCharCodeOfElementAtIndex(element: string, index: number, numPasses: number): number;
    /**
     * Number of passes is determined by the length of the longest element in the array.
     * For integers, this log10(num), and for strings, this would be the length of the string.
     */
    determineNumPasses(array: T[]): number;
    /**
     * @param array
     */
    getLengthOfLongestElement(array: any[]): number;
    /**
     * @param array
     */
    isArrayOfNumbers(array: T[]): boolean;
    /**
     * @param numBuckets
     */
    createBuckets(numBuckets: number): T[][];
    /**
     * @param element
     */
    isNumber(element: T): boolean;
}

declare class SelectionSort<T> extends Sort<T> {
    sort(originalArray: T[]): T[];
}

declare class ShellSort<T> extends Sort<T> {
    sort(originalArray: T[]): T[];
}

/**
 * Picks the random item based on its weight.
 * The items with higher weight will be picked more often (with a higher probability).
 *
 * For example:
 * - items = ['banana', 'orange', 'apple']
 * - weights = [0, 0.2, 0.8]
 * - weightedRandom(items, weights) in 80% of cases will return 'apple', in 20% of cases will return
 * 'orange' and it will never return 'banana' (because probability of picking the banana is 0%)
 *
 * @param items
 * @param weights
 */
declare function weightedRandom<T>(items: T[], weights: number[]): {
    item: T;
    index: number;
};

/**
 * @param a
 * @param b
 */
declare function hammingDistance(a: string, b: string): number;

/**
 * @param text
 * @param word
 */
declare function knuthMorrisPratt(text: string, word: string): number;

/**
 * 最长公共子串
 *
 * @param string1
 * @param string2
 */
declare function longestCommonSubstring(string1: string, string2: string): string;

/**
 * 是否回文
 *
 * @param string
 */
declare function isPalindrome(string: string): boolean;

/**
 * 卡普-拉宾算法（Karp–Rabin algorithm）
 *
 * @param text - Text that may contain the searchable word.
 * @param word - Word that is being searched in text.
 * @return Position of the word in text.
 */
declare function rabinKarp(text: string, word: string): number;

/**
 * Dynamic programming approach.
 *
 * @param string
 * @param pattern
 */
declare function regularExpressionMatching(string: string, pattern: string): boolean;

/**
 * @param text
 * @param word
 */
declare function zAlgorithm(text: string, word: string): number[];

/**
 * 哈希表（散列表）
 */
declare class HashTable<T> {
    private keys;
    buckets: LinkedList<{
        key: string;
        value: T;
    }>[];
    /**
     * 构建哈希表
     *
     * @param hashTableSize 哈希表尺寸
     */
    constructor(hashTableSize?: number);
    /**
     * Converts key string to hash number.
     * 将字符串键转换为哈希数。
     *
     * @param key 字符串键
     * @return 哈希数
     */
    hash(key: string): number;
    /**
     * 设置值
     *
     * @param key 键
     * @param value 值
     */
    set(key: string, value: any): void;
    /**
     * 删除指定键以及对于值
     *
     * @param key 需要删除的键
     * @return 被删除的结点
     */
    delete(key: string): LinkedListNode<{
        key: string;
        value: T;
    }>;
    /**
     * 获取与键对应的值
     *
     * @param key 键
     * @return
     */
    get(key: string): T;
    /**
     * 是否拥有键
     *
     * @param key 键
     * @return 是否拥有键
     */
    has(key: string): any;
    /**
     * 获取键列表
     *
     * @return 键列表
     */
    getKeys(): string[];
    /**
     * Gets the list of all the stored values in the hash table.
     *
     * @return 值列表
     */
    getValues(): T[];
}

declare class BinaryTreeNode<T> {
    /**
     * 左结点
     */
    left: BinaryTreeNode<T>;
    /**
     * 右结点
     */
    right: BinaryTreeNode<T>;
    /**
     * 父结点
     */
    parent: BinaryTreeNode<T>;
    /**
     * 结点值
     */
    value: T;
    /**
     * 结点比较器
     */
    nodeComparator: Comparator<BinaryTreeNode<T>>;
    meta: HashTable<string>;
    /**
     * 构建二叉树结点
     *
     * @param value node value.
     */
    constructor(value?: T);
    /**
     * 左结点高度
     */
    get leftHeight(): any;
    /**
     * 右结点高度
     */
    get rightHeight(): any;
    /**
     * 高度
     */
    get height(): number;
    /**
     * 平衡系数
     */
    get balanceFactor(): number;
    /**
     * Get parent's sibling if it exists.
     *
     * @return 叔伯结点
     */
    get uncle(): BinaryTreeNode<T>;
    /**
     * 设置结点值
     *
     * @param value 结点新值
     */
    setValue(value: T): this;
    /**
     * 设置左结点
     *
     * @param node 结点
     */
    setLeft(node: BinaryTreeNode<T>): this;
    /**
     * 设置右结点
     *
     * @param node 结点
     */
    setRight(node: BinaryTreeNode<T>): this;
    /**
     * 移除子结点
     *
     * @param nodeToRemove 需要移除的子结点
     * @return 是否成功移除
     */
    removeChild(nodeToRemove: BinaryTreeNode<T>): boolean;
    /**
     * 替换节点
     *
     * @param nodeToReplace 被替换的节点
     * @param replacementNode 替换后的节点
     * @return 是否替换成功
     */
    replaceChild(nodeToReplace: any, replacementNode: any): boolean;
    /**
     * 拷贝节点
     *
     * @param sourceNode 源节点
     * @param targetNode 目标节点
     */
    static copyNode<T>(sourceNode: BinaryTreeNode<T>, targetNode: BinaryTreeNode<T>): void;
    /**
     * 左序深度遍历
     */
    traverseInOrder(): any[];
    /**
     * 转换为字符串
     */
    toString(): string;
}

interface Callbacks<T> {
    /**
     * Determines whether BFS should traverse from the node to its child.
     */
    allowTraversal?: (node: BinaryTreeNode<T>, child: BinaryTreeNode<T>) => boolean;
    /**
     * Called when BFS enters the node.
     */
    enterNode?: (node: BinaryTreeNode<T>) => void;
    /**
     * Called when BFS leaves the node.
     */
    leaveNode?: (node: BinaryTreeNode<T>) => void;
}
/**
 * 广度优先搜索算法（英语：Breadth-First Search，缩写为BFS）
 *
 * @param rootNode
 * @param originalCallbacks
 */
declare function breadthFirstSearch1<T>(rootNode: BinaryTreeNode<T>, originalCallbacks?: Callbacks<T>): void;

interface TraversalCallbacks<T> {
    /**
     * Determines whether DFS should traverse from the node to its child.
     */
    allowTraversal?: (node: BinaryTreeNode<T>, child: BinaryTreeNode<T>) => boolean;
    /**
     * Called when DFS enters the node.
     */
    enterNode?: (node: BinaryTreeNode<T>) => void;
    /**
     * Called when DFS leaves the node.
     */
    leaveNode?: (node: BinaryTreeNode<T>) => void;
}
/**
 * Recursive depth-first-search traversal for binary.
 *
 * @param node binary tree node that we will start traversal from.
 * @param callbacks the object that contains traversal callbacks.
 */
declare function depthFirstSearchRecursive<T>(node: BinaryTreeNode<T>, callbacks: TraversalCallbacks<T>): void;
/**
 * Perform depth-first-search traversal of the rootNode.
 * For every traversal step call "allowTraversal", "enterNode" and "leaveNode" callbacks.
 * See TraversalCallbacks type definition for more details about the shape of callbacks object.
 *
 * @param rootNode The node from which we start traversing.
 * @param callbacks Traversal callbacks.
 */
declare function depthFirstSearch1<T>(rootNode: BinaryTreeNode<T>, callbacks?: TraversalCallbacks<T>): void;

/**
 * Finds the maximum profit from selling and buying the stocks.
 * ACCUMULATOR APPROACH.
 *
 * @param prices Array of stock prices, i.e. [7, 6, 4, 3, 1]
 * @param visit Visiting callback to calculate the number of iterations.
 * @return The maximum profit
 */
declare function accumulatorBestTimeToBuySellStocks(prices: number[], visit?: () => void): number;

/**
 * Finds the maximum profit from selling and buying the stocks.
 * DYNAMIC PROGRAMMING APPROACH.
 *
 * @param prices Array of stock prices, i.e. [7, 6, 4, 3, 1]
 * @param visit Visiting callback to calculate the number of iterations.
 * @return The maximum profit
 */
declare function dpBestTimeToBuySellStocks(prices: number[], visit?: () => void): number;

/**
 * Finds the maximum profit from selling and buying the stocks.
 * DIVIDE & CONQUER APPROACH.
 *
 * @param prices Array of stock prices, i.e. [7, 6, 4, 3, 1]
 * @param visit Visiting callback to calculate the number of iterations.
 * @return The maximum profit
 */
declare function dqBestTimeToBuySellStocks(prices: number[], visit?: () => void): number;

/**
 * Finds the maximum profit from selling and buying the stocks.
 * PEAK VALLEY APPROACH.
 *
 * @param prices Array of stock prices, i.e. [7, 6, 4, 3, 1]
 * @param visit Visiting callback to calculate the number of iterations.
 * @return The maximum profit
 */
declare function peakvalleyBestTimeToBuySellStocks(prices: number[], visit?: () => void): number;

/**
 * 栈
 *
 * 后进先出
 */
declare class Stack<T> {
    linkedList: LinkedList<T>;
    constructor();
    /**
     * 是否为空
     */
    isEmpty(): boolean;
    /**
     * 查看第一个元素值
     */
    peek(): T;
    /**
     * 入栈
     *
     * @param value 元素值
     */
    push(value: T): void;
    /**
     * 出栈
     */
    pop(): T;
    /**
     * 转换为数组
     */
    toArray(): T[];
    /**
     * 转换为字符串
     *
     * @param callback 值输出为字符串函数
     */
    toString(callback?: (value: T) => string): string;
}

declare function hanoiTower<T>({ numberOfDiscs, moveCallback, fromPole, withPole, toPole, }: {
    numberOfDiscs: number;
    moveCallback: (disc: T, fromPole: T[], toPole: T[]) => void;
    fromPole?: Stack<T>;
    withPole?: Stack<T>;
    toPole?: Stack<T>;
}): void;

/**
 * BACKTRACKING approach of solving Jump Game.
 *
 * This is the inefficient solution where we try every single jump
 * pattern that takes us from the first position to the last.
 * We start from the first position and jump to every index that
 * is reachable. We repeat the process until last index is reached.
 * When stuck, backtrack.
 *
 * @param numbers array of possible jump length.
 * @param startIndex index from where we start jumping.
 * @param currentJumps current jumps path.
 */
declare function backtrackingJumpGame(numbers: number[], startIndex?: number, currentJumps?: number[]): boolean;

/**
 * DYNAMIC PROGRAMMING BOTTOM-UP approach of solving Jump Game.
 *
 * This comes out as an optimisation of DYNAMIC PROGRAMMING TOP-DOWN approach.
 *
 * The observation to make here is that we only ever jump to the right.
 * This means that if we start from the right of the array, every time we
 * will query a position to our right, that position has already be
 * determined as being GOOD or BAD. This means we don't need to recurse
 * anymore, as we will always hit the memo table.
 *
 * We call a position in the array a "good" one if starting at that
 * position, we can reach the last index. Otherwise, that index
 * is called a "bad" one.
 *
 * @param numbers array of possible jump length.
 */
declare function dpBottomUpJumpGame(numbers: number[]): boolean;

/**
 * DYNAMIC PROGRAMMING TOP-DOWN approach of solving Jump Game.
 *
 * This comes out as an optimisation of BACKTRACKING approach.
 *
 * It relies on the observation that once we determine that a certain
 * index is good / bad, this result will never change. This means that
 * we can store the result and not need to recompute it every time.
 *
 * We call a position in the array a "good" one if starting at that
 * position, we can reach the last index. Otherwise, that index
 * is called a "bad" one.
 *
 * @param numbers array of possible jump length.
 * @param startIndex index from where we start jumping.
 * @param currentJumps current jumps path.
 * @param cellsGoodness holds information about whether cell is "good" or "bad"
 */
declare function dpTopDownJumpGame(numbers: number[], startIndex?: number, currentJumps?: number[], cellsGoodness?: boolean[]): boolean;

/**
 * GREEDY approach of solving Jump Game.
 *
 * This comes out as an optimisation of DYNAMIC PROGRAMMING BOTTOM_UP approach.
 *
 * Once we have our code in the bottom-up state, we can make one final,
 * important observation. From a given position, when we try to see if
 * we can jump to a GOOD position, we only ever use one - the first one.
 * In other words, the left-most one. If we keep track of this left-most
 * GOOD position as a separate variable, we can avoid searching for it
 * in the array. Not only that, but we can stop using the array altogether.
 *
 * We call a position in the array a "good" one if starting at that
 * position, we can reach the last index. Otherwise, that index
 * is called a "bad" one.
 *
 * @param numbers array of possible jump length.
 */
declare function greedyJumpGame(numbers: number[]): boolean;

/**
 * @param chessboardSize
 */
declare function knightTour(chessboardSize: number): number[][];

/**
 * Class that represents queen position on the chessboard.
 */
declare class QueenPosition {
    rowIndex: number;
    columnIndex: number;
    /**
     * @param rowIndex
     * @param columnIndex
     */
    constructor(rowIndex: number, columnIndex: number);
    get leftDiagonal(): number;
    get rightDiagonal(): number;
    toString(): string;
}

/**
 * @param queensCount
 */
declare function nQueens(queensCount: number): QueenPosition[][];

/**
 * BRUTE FORCE approach of solving Trapping Rain Water problem.
 *
 * @param terraces
 */
declare function bfRainTerraces(terraces: number[]): number;

/**
 * DYNAMIC PROGRAMMING approach of solving Trapping Rain Water problem.
 *
 * @param terraces
 */
declare function dpRainTerraces(terraces: number[]): number;

/**
 * Recursive Staircase Problem (Brute Force Solution).
 *
 * @param stairsNum Number of stairs to climb on.
 * @return Number of ways to climb a staircase.
 */
declare function recursiveStaircaseBF(stairsNum: number): number;

/**
 * Recursive Staircase Problem (Dynamic Programming Solution).
 *
 * @param stairsNum Number of stairs to climb on.
 * @return Number of ways to climb a staircase.
 */
declare function recursiveStaircaseDP(stairsNum: number): number;

/**
 * Recursive Staircase Problem (Iterative Solution).
 *
 * @param stairsNum Number of stairs to climb on.
 * @return Number of ways to climb a staircase.
 */
declare function recursiveStaircaseIT(stairsNum: number): number;

/**
 * Recursive Staircase Problem (Recursive Solution With Memoization).
 *
 * @param totalStairs Number of stairs to climb on.
 * @return Number of ways to climb a staircase.
 */
declare function recursiveStaircaseMEM(totalStairs: number): number;

/**
 * @param originalMatrix
 */
declare function squareMatrixRotation<T>(originalMatrix: T[][]): T[][];

/**
 * BACKTRACKING approach of solving Unique Paths problem.
 *
 * @param width Width of the board.
 * @param height Height of the board.
 * @param steps The steps that have been already made.
 * @param uniqueSteps - Total number of unique steps.
 * @return Number of unique paths.
 */
declare function btUniquePaths(width: number, height: number, steps?: number[][], uniqueSteps?: number): number;

/**
 * DYNAMIC PROGRAMMING approach of solving Unique Paths problem.
 *
 * @param width Width of the board.
 * @param height Height of the board.
 * @return Number of unique paths.
 */
declare function dpUniquePaths(width: number, height: number): number;

/**
 * @param width
 * @param height
 */
declare function uniquePaths(width: number, height: number): number;

/**
 * 布隆过滤器
 */
declare class BloomFilter {
    private size;
    private storage;
    /**
     * @param size the size of the storage.
     */
    constructor(size?: number);
    /**
     * 插入
     *
     * @param item 元素
     */
    insert(item: string): void;
    /**
     * 可能包含
     *
     * @param item 元素
     * @return
     */
    mayContain(item: string): boolean;
    /**
     * Creates the data store for our filter.
     * We use this method to generate the store in order to
     * encapsulate the data itself and only provide access
     * to the necessary methods.
     *
     * @param size 尺寸
     * @return
     */
    createStore(size: number): {
        getValue(index: number): any;
        setValue(index: number): void;
    };
    /**
     * 计算哈希值1
     *
     * @param item 元素
     * @return 哈希值1
     */
    hash1(item: string): number;
    /**
     * 计算哈希值2
     *
     * @param item 元素
     * @return 哈希值2
     */
    hash2(item: string): number;
    /**
     * 计算哈希值3
     *
     * @param item 元素
     * @return 哈希值3
     */
    hash3(item: string): number;
    /**
     * Runs all 3 hash functions on the input and returns an array of results.
     * 获取3个哈希值组成的数组
     *
     * @param item 元素
     * @return 3个哈希值组成的数组
     */
    getHashValues(item: string): number[];
}

/**
 * 并查集
 *
 * 并查集是一种树型的数据结构，用于处理一些不交集（Disjoint Sets）的合并及查询问题。
 *
 * @see https://github.com/trekhleb/javascript-algorithms/blob/master/src/data-structures/disjoint-set/DisjointSet.js
 * @see https://en.wikipedia.org/wiki/Disjoint-set_data_structure
 * @see https://www.youtube.com/watch?v=wU6udHRIkcc&index=14&t=0s&list=PLLXdhg_r2hKA7DPDsunoDZ-Z769jWn4R8
 */
declare class DisjointSet<T> {
    private items;
    /**
     * 计算键值函数
     */
    private keyCallback;
    /**
     * 构建 并查集
     *
     * @param keyCallback 计算键值函数
     */
    constructor(keyCallback?: (value: T) => string);
    /**
     * 创建集合
     *
     * @param itemValue 结点值
     * @return 创建的集合
     */
    makeSet(itemValue: T): this;
    /**
     * Find set representation node.
     *
     * @param itemValue 结点值
     */
    find(itemValue: T): string;
    /**
     * Union by rank.
     *
     * @param valueA 值A
     * @param valueB 值B
     * @return 自身
     */
    union(valueA: T, valueB: T): this;
    /**
     * 判断两个值是否在相同集合中
     *
     * @param valueA 值A
     * @param valueB 值B
     */
    inSameSet(valueA: T, valueB: T): boolean;
}

/**
 * 并查集结点
 */
declare class DisjointSetItem<T> {
    /**
     * 值
     */
    private value;
    /**
     * 计算键值函数
     */
    private keyCallback;
    /**
     * 父结点
     */
    private parent;
    /**
     * 子结点
     */
    private children;
    /**
     * 构建 并查集 项
     *
     * @param value 值
     * @param keyCallback 计算键值函数
     */
    constructor(value: T, keyCallback?: (value: T) => string);
    /**
     * 获取键值
     */
    getKey(): string;
    /**
     * 获取根结点
     */
    getRoot(): DisjointSetItem<T>;
    /**
     * 是否为根结点
     */
    isRoot(): boolean;
    /**
     * Rank basically means the number of all ancestors.
     *
     * @return 所有子孙结点数量
     */
    getRank(): number;
    /**
     * 获取子结点列表
     *
     * @return 子结点列表
     */
    getChildren(): DisjointSetItem<T>[];
    /**
     * 设置父结点
     *
     * @param parentItem 父结点
     * @param forceSettingParentChild 强制这是父子关系
     * @return 自身
     */
    setParent(parentItem: DisjointSetItem<T>, forceSettingParentChild?: boolean): this;
    /**
     * 添加子结点
     *
     * @param childItem 子结点
     * @return 自身
     */
    addChild(childItem: DisjointSetItem<T>): this;
}

/**
 * 双向链接结点
 */
declare class DoublyLinkedListNode<T> {
    /**
     * 值
     */
    value: T;
    /**
     * 上一个结点
     */
    previous: DoublyLinkedListNode<T>;
    /**
     * 下一个结点
     */
    next: DoublyLinkedListNode<T>;
    constructor(value: T, next?: DoublyLinkedListNode<T>, previous?: DoublyLinkedListNode<T>);
    toString(callback?: (value: T) => string): string;
}

/**
 * 双向链表
 */
declare class DoublyLinkedList<T> {
    /**
     * 表头
     */
    head: DoublyLinkedListNode<T>;
    /**
     * 表尾
     */
    tail: DoublyLinkedListNode<T>;
    /**
     * 比较器
     */
    private compare;
    /**
     * 构建双向链表
     *
     * @param comparatorFunction 比较函数
     */
    constructor(comparatorFunction?: CompareFunction<T>);
    /**
     * 在头部新增结点
     *
     * @param value 新增结点值
     * @return 新增的结点
     */
    prepend(value: T): this;
    /**
     * 在尾部新增结点
     *
     * @param value 新增结点值
     * @return 新增的结点
     */
    append(value: T): this;
    /**
     * 删除链表中第一个与指定值相等的结点
     *
     * @param value 结点值
     * @return 删除的结点
     */
    delete(value: T): any;
    /**
     * 查找与结点值相等的结点
     *
     * @return 找到的第一个结点
     */
    find({ value, callback }: {
        /**
         * 查找的值
         */
        value?: T;
        /**
         * 比较函数
         */
        callback?: (value: T) => boolean;
    }): DoublyLinkedListNode<T>;
    /**
     * 删除表尾
     *
     * @return 被删除的结点
     */
    deleteTail(): DoublyLinkedListNode<T>;
    /**
     * 删除表头
     *
     * @return 被删除的表头
     */
    deleteHead(): DoublyLinkedListNode<T>;
    /**
     * 转换为数组
     *
     * @return 链表结点组成的函数
     */
    toArray(): DoublyLinkedListNode<T>[];
    /**
     * 从数组中初始化链表
     *
     * @param values - Array of values that need to be converted to linked list.
     * @return 初始化后的链表
     */
    fromArray(values: T[]): this;
    /**
     * 转换为字符串
     *
     * @param callback
     * @return 字符串
     */
    toString(callback?: (value: T) => string): string;
    /**
     * Reverse a linked list.
     *
     * @returns 反转后的链表
     */
    reverse(): this;
}

/**
 * 堆
 *
 * Parent class for Min and Max Heaps.
 */
declare class Heap<T> {
    /**
     * 堆的数组表示。
     */
    private heapContainer;
    /**
     * 比较器
     */
    protected compare: Comparator<T>;
    /**
     * 构建链表
     *
     * @param comparatorFunction 比较函数
     */
    constructor(comparatorFunction?: CompareFunction<T>);
    /**
     * 获取左边子结点索引
     *
     * @param parentIndex 父结点索引
     * @return 左边子结点索引
     */
    getLeftChildIndex(parentIndex: number): number;
    /**
     * 获取右边子结点索引
     *
     * @param parentIndex 父结点索引
     * @return 右边子结点索引
     */
    getRightChildIndex(parentIndex: number): number;
    /**
     * 获取父结点索引
     *
     * @param childIndex 子结点索引
     * @return 父结点索引
     */
    getParentIndex(childIndex: number): number;
    /**
     * 是否有父结点
     *
     * @param childIndex 子结点索引
     * @return 是否有父结点
     */
    hasParent(childIndex: number): boolean;
    /**
     * 是否有左子结点
     *
     * @param parentIndex 父结点索引
     * @return 是否有左子结点
     */
    hasLeftChild(parentIndex: number): boolean;
    /**
     * 是否有右子结点
     *
     * @param parentIndex 父结点索引
     * @return 是否有右子结点
     */
    hasRightChild(parentIndex: number): boolean;
    /**
     * 获取左结点
     *
     * @param parentIndex 父结点索引
     * @return 左结点
     */
    leftChild(parentIndex: number): T;
    /**
     * 获取右结点
     *
     * @param parentIndex 父结点索引
     * @return 右结点
     */
    rightChild(parentIndex: number): T;
    /**
     * 获取父结点
     *
     * @param childIndex 子结点索引
     * @return 父结点
     */
    parent(childIndex: number): T;
    /**
     * 交换两个结点数据
     *
     * @param indexOne 索引1
     * @param indexTwo 索引2
     */
    swap(indexOne: number, indexTwo: number): void;
    /**
     * 查看堆顶数据
     *
     * @return 堆顶数据
     */
    peek(): T;
    /**
     * 取出堆顶元素
     *
     * @return 堆顶元素
     */
    poll(): T;
    /**
     * 新增元素
     *
     * @param item 新元素
     * @return 自身
     */
    add(item: T): this;
    /**
     * 移除所有指定元素
     *
     * @param item 被移除的元素
     * @param comparator 比较器
     * @return 自身
     */
    remove(item: T, comparator?: Comparator<T>): this;
    /**
     * 查找元素所在的所有索引
     *
     * @param item 查找的元素
     * @param comparator 比较器
     * @return 查找的元素所在索引了吧
     */
    find(item: T, comparator?: Comparator<T>): number[];
    /**
     * 是否为空
     */
    isEmpty(): boolean;
    /**
     * 转换为字符串
     */
    toString(): string;
    /**
     * 堆冒泡
     *
     * @param customStartIndex 堆冒泡起始索引
     */
    heapifyUp(customStartIndex?: number): void;
    /**
     * 堆下沉
     *
     * @param customStartIndex 堆下沉起始索引
     */
    heapifyDown(customStartIndex?: number): void;
    /**
     * Checks if pair of heap elements is in correct order.
     * For MinHeap the first element must be always smaller or equal.
     * For MaxHeap the first element must be always bigger or equal.
     *
     * @param firstElement 第一个元素
     * @param secondElement 第二个元素
     * @return {boolean}
     */
    pairIsInCorrectOrder(firstElement: T, secondElement: T): boolean;
}

/**
 * 最大堆
 *
 * 所有父结点都大于子结点
 */
declare class MaxHeap<T> extends Heap<T> {
    /**
     * Checks if pair of heap elements is in correct order.
     * For MinHeap the first element must be always smaller or equal.
     * For MaxHeap the first element must be always bigger or equal.
     *
     * 检查堆元素对的顺序是否正确。
     * 对于MinHeap，第一个元素必须总是小于等于。
     * 对于MaxHeap，第一个元素必须总是大于或等于。
     *
     * @param firstElement 第一个元素
     * @param secondElement 第二个元素
     * @return 堆元素对的顺序是否正确
     */
    pairIsInCorrectOrder(firstElement: T, secondElement: T): boolean;
}

/**
 * 最小堆
 *
 * 所有父结点都小于子结点
 */
declare class MinHeap<T> extends Heap<T> {
    /**
     * Checks if pair of heap elements is in correct order.
     * For MinHeap the first element must be always smaller or equal.
     * For MaxHeap the first element must be always bigger or equal.
     *
     * 检查堆元素对的顺序是否正确。
     * 对于MinHeap，第一个元素必须总是小于等于。
     * 对于MaxHeap，第一个元素必须总是大于或等于。
     *
     * @param firstElement 第一个元素
     * @param secondElement 第二个元素
     * @return 堆元素对的顺序是否正确
     */
    pairIsInCorrectOrder(firstElement: T, secondElement: T): boolean;
}

/**
 * 优先队列
 *
 * It is the same as min heap except that when comparing two elements
 * we take into account its priority instead of the element's value.
 */
declare class PriorityQueue<T> extends MinHeap<T> {
    private priorities;
    constructor();
    /**
     * Add item to the priority queue.
     *
     * @param item item we're going to add to the queue.
     * @param priority items priority.
     */
    add(item: T, priority?: number): this;
    /**
     * Remove item from priority queue.
     *
     * @param item item we're going to remove.
     * @param customFindingComparator custom function for finding the item to remove
     */
    remove(item: T, customFindingComparator: Comparator<T>): this;
    /**
     * Change priority of the item in a queue.
     *
     * @param item item we're going to re-prioritize.
     * @param priority new item's priority.
     */
    changePriority(item: T, priority: number): this;
    /**
     * Find item by ite value.
     *
     * @param item
     */
    findByValue(item: T): number[];
    /**
     * Check if item already exists in a queue.
     *
     * @param item 元素
     */
    hasValue(item: T): boolean;
    /**
     * Compares priorities of two items.
     *
     * @param a 元素a
     * @param b 元素b
     */
    comparePriority(a: T, b: T): 0 | 1 | -1;
    /**
     * Compares values of two items.
     * @param a 元素a
     * @param b 元素b
     */
    compareValue(a: T, b: T): 0 | 1 | -1;
}

/**
 * 队列，只能从后面进，前面出
 * 使用单向链表实现
 */
declare class Queue<T> {
    linkedList: LinkedList<T>;
    constructor();
    /**
     * 是否为空
     */
    isEmpty(): boolean;
    /**
     * Read the element at the front of the queue without removing it.
     *
     * 读取队列前面的元素，但不删除它。
     */
    peek(): T;
    /**
     * 入队
     *
     * 在队列的末尾(链表的尾部)添加一个新元素。
     * 这个元素将在它前面的所有元素之后被处理。
     *
     * Add a new element to the end of the queue (the tail of the linked list).
     * This element will be processed after all elements ahead of it.
     *
     * @param value 元素值
     */
    enqueue(value: T): void;
    /**
     * 出队
     *
     * 删除队列前面的元素(链表的头)。如果队列为空，则返回null。
     *
     * Remove the element at the front of the queue (the head of the linked list).
     * If the queue is empty, return null.
     *
     * @return 出队的元素
     */
    dequeue(): T;
    /**
     * 转换为字符串
     *
     * @param callback 值输出为字符串函数
     */
    toString(callback?: (value: T) => string): string;
}

/**
 * 二叉查找树结点
 */
declare class BinarySearchTreeNode<T> extends BinaryTreeNode<T> {
    /**
     * 左结点
     */
    left: BinarySearchTreeNode<T>;
    /**
     * 右结点
     */
    right: BinarySearchTreeNode<T>;
    /**
     * 父结点
     */
    parent: BinarySearchTreeNode<T>;
    /**
     * 比较函数
     */
    private compareFunction;
    /**
     * 结点值比较器
     */
    private nodeValueComparator;
    /**
     * 构建二叉查找树结点
     *
     * @param value node value.
     * @param compareFunction comparator function for node values.
     */
    constructor(value?: T, compareFunction?: CompareFunction<T>);
    /**
     * 插入值
     *
     * @param value 值
     */
    insert(value: T): BinarySearchTreeNode<T>;
    /**
     * 查找结点
     *
     * @param value 值
     * @return {BinarySearchTreeNode}
     */
    find(value: T): BinarySearchTreeNode<T>;
    /**
     * 是否包含指定值
     *
     * @param value 指定值
     */
    contains(value: T): boolean;
    /**
     * 移除指定值
     *
     * @param  value 指定值
     * @return 是否移除成功
     */
    remove(value: T): boolean;
    /**
     * 查找最小值
     */
    findMin(): BinarySearchTreeNode<T>;
}

/**
 * 二叉查找树
 *
 * 二叉查找树（英语：Binary Search Tree），也称为二叉搜索树、有序二叉树（ordered binary tree）或排序二叉树（sorted binary tree），是指一棵空树或者具有下列性质的二叉树：
 *
 * 1. 若任意节点的左子树不空，则左子树上所有节点的值均小于它的根节点的值；
 * 2. 若任意节点的右子树不空，则右子树上所有节点的值均大于它的根节点的值；
 * 3. 任意节点的左、右子树也分别为二叉查找树；
 * 4. 没有键值相等的节点。
 */
declare class BinarySearchTree<T> {
    /**
     * 根结点
     */
    root: BinarySearchTreeNode<T>;
    /**
     * 结点比较器
     */
    nodeComparator: Comparator<BinaryTreeNode<T>>;
    /**
     * 构建 二叉查找树
     *
     * @param nodeValueCompareFunction 结点值比较器
     */
    constructor(nodeValueCompareFunction?: CompareFunction<T>);
    /**
     * 插入值
     *
     * @param value 值
     */
    insert(value: T): BinarySearchTreeNode<T>;
    /**
     * 是否包含指定值
     *
     * @param value 值
     */
    contains(value: T): boolean;
    /**
     * 移除指定值
     *
     * @param value 值
     */
    remove(value: T): boolean;
    /**
     * 转换为字符串
     */
    toString(): string;
}

declare class AvlTree<T> extends BinarySearchTree<T> {
    /**
     * @param value
     */
    insert(value: T): BinarySearchTreeNode<T>;
    /**
     * @param value
     */
    remove(value: T): boolean;
    /**
     * 平衡结点
     *
     * @param node 结点
     */
    balance(node: BinarySearchTreeNode<T>): void;
    /**
     * @param  rootNode
     */
    rotateLeftLeft(rootNode: BinarySearchTreeNode<T>): void;
    /**
     * @param  rootNode
     */
    rotateLeftRight(rootNode: BinarySearchTreeNode<T>): void;
    /**
     * @param  rootNode
     */
    rotateRightLeft(rootNode: BinarySearchTreeNode<T>): void;
    /**
     * @param rootNode
     */
    rotateRightRight(rootNode: BinarySearchTreeNode<T>): void;
}

/**
 * 树状数组（Fenwick树、Binary Indexed Tree）
 *
 * @see https://zh.wikipedia.org/wiki/%E6%A0%91%E7%8A%B6%E6%95%B0%E7%BB%84
 */
declare class FenwickTree {
    arraySize: number;
    treeArray: number[];
    /**
     * Constructor creates empty fenwick tree of size 'arraySize',
     * however, array size is size+1, because index is 1-based.
     *
     * @param arraySize
     */
    constructor(arraySize: number);
    /**
     * Adds value to existing value at position.
     *
     * @param position
     * @param value
     */
    increase(position: number, value: number): this;
    /**
     * Query sum from index 1 to position.
     *
     * @param position
     */
    query(position: number): number;
    /**
     * Query sum from index leftIndex to rightIndex.
     *
     * @param leftIndex
     * @param rightIndex
     */
    queryRange(leftIndex: number, rightIndex: number): number;
}

/**
 * 红黑树
 */
declare class RedBlackTree<T> extends BinarySearchTree<T> {
    /**
     * @param value
     */
    insert(value: T): BinarySearchTreeNode<T>;
    /**
     * @param value
     */
    remove(value: T): boolean;
    /**
     * @param node
     */
    balance(node: BinarySearchTreeNode<T>): void;
    /**
     * Left Left Case (p is left child of g and x is left child of p)
     * @param grandParentNode
     */
    leftLeftRotation(grandParentNode: BinarySearchTreeNode<T>): BinarySearchTreeNode<T>;
    /**
     * Left Right Case (p is left child of g and x is right child of p)
     *
     * @param grandParentNode
     */
    leftRightRotation(grandParentNode: BinarySearchTreeNode<T>): BinarySearchTreeNode<T>;
    /**
     * Right Right Case (p is right child of g and x is right child of p)
     *
     * @param grandParentNode
     */
    rightRightRotation(grandParentNode: BinarySearchTreeNode<T>): BinarySearchTreeNode<T>;
    /**
     * Right Left Case (p is right child of g and x is left child of p)
     *
     * @param grandParentNode
     */
    rightLeftRotation(grandParentNode: BinarySearchTreeNode<T>): BinarySearchTreeNode<T>;
    /**
     * @param node
     */
    makeNodeRed(node: BinarySearchTreeNode<T>): BinarySearchTreeNode<T>;
    /**
     * @param node
     */
    makeNodeBlack(node: BinarySearchTreeNode<T>): BinarySearchTreeNode<T>;
    /**
     * @param node
     */
    isNodeRed(node: BinarySearchTreeNode<T>): boolean;
    /**
     * @param node
     */
    isNodeBlack(node: BinarySearchTreeNode<T>): boolean;
    /**
     * @param node
     */
    isNodeColored(node: BinarySearchTreeNode<T>): boolean;
    /**
     * @param firstNode
     * @param secondNode
     */
    swapNodeColors(firstNode: BinarySearchTreeNode<T>, secondNode: BinarySearchTreeNode<T>): void;
}

/**
 * 线段树
 */
declare class SegmentTree {
    inputArray: number[];
    /**
     * binary function (i.e. sum, min)
     */
    operation: (a: number, b: number) => number;
    /**
     * operation fallback value (i.e. 0 for sum, Infinity for min)
     */
    operationFallback: number;
    segmentTree: number[];
    /**
     * @param inputArray
     * @param operation binary function (i.e. sum, min)
     * @param operationFallback operation fallback value (i.e. 0 for sum, Infinity for min)
     */
    constructor(inputArray: number[], operation: any, operationFallback: number);
    /**
     *
     * @param inputArray
     */
    initSegmentTree(inputArray: number[]): number[];
    /**
     * Build segment tree.
     */
    buildSegmentTree(): void;
    /**
     * Build segment tree recursively.
     *
     * @param leftInputIndex
     * @param rightInputIndex
     * @param position
     */
    buildTreeRecursively(leftInputIndex: number, rightInputIndex: number, position: number): void;
    /**
     * Do range query on segment tree in context of this.operation function.
     *
     * @param queryLeftIndex
     * @param queryRightIndex
     */
    rangeQuery(queryLeftIndex: number, queryRightIndex: number): number;
    /**
     * Do range query on segment tree recursively in context of this.operation function.
     *
     * @param queryLeftIndex left index of the query
     * @param queryRightIndex right index of the query
     * @param leftIndex left index of input array segment
     * @param rightIndex right index of input array segment
     * @param position root position in binary tree
     */
    rangeQueryRecursive(queryLeftIndex: number, queryRightIndex: number, leftIndex: number, rightIndex: number, position: number): number;
    /**
     * Left child index.
     *
     * @param parentIndex
     */
    getLeftChildIndex(parentIndex: number): number;
    /**
     * Right child index.
     *
     * @param parentIndex
     */
    getRightChildIndex(parentIndex: number): number;
}

declare class TrieNode<T> {
    /**
     * 字符
     */
    character: string;
    /**
     * 是否完整单词
     */
    isCompleteWord: boolean;
    /**
     * 子结点哈希表
     */
    children: HashTable<TrieNode<T>>;
    /**
     * @param character 字符
     * @param isCompleteWord 是否完整单词
     */
    constructor(character: string, isCompleteWord?: boolean);
    /**
     * 获取子结点
     *
     * @param character 字符
     */
    getChild(character: string): TrieNode<T>;
    /**
     * 添加子结点
     *
     * @param character 字符
     * @param isCompleteWord 字符完整单词
     * @return 自身
     */
    addChild(character: string, isCompleteWord?: boolean): TrieNode<T>;
    /**
     * 移除结点
     *
     * @param character 字符
     * @return 自身
     */
    removeChild(character: string): this;
    /**
     * 是否有指定子结点
     *
     * @param character 字符
     */
    hasChild(character: string): any;
    /**
     * Check whether current TrieNode has children or not.
     */
    hasChildren(): boolean;
    /**
     *
     */
    suggestChildren(): string[];
    /**
     * 转换为字符串
     */
    toString(): string;
}

/**
 * 字典树
 */
declare class Trie {
    head: TrieNode<string>;
    constructor();
    /**
     * 添加单词
     *
     * @param word 单词
     * @return 自身
     */
    addWord(word: string): this;
    /**
     * 删除单词
     *
     * @param word 单词
     * @return 自身
     */
    deleteWord(word: string): this;
    /**
     *
     * @param word 单词
     */
    suggestNextCharacters(word: string): string[];
    /**
     * Check if complete word exists in Trie.
     *
     * @param word 单词
     */
    doesWordExist(word: string): boolean;
    /**
     *
     * @param word 单词
     */
    getLastCharacterNode(word: string): TrieNode<string>;
}

export { AvlTree, BinarySearchTree, BinarySearchTreeNode, BinaryTreeNode, BloomFilter, BubbleSort, Comparator, CompareFunction, ComplexNumber, CountingSort, DisjointSet, DisjointSetItem, DoublyLinkedList, DoublyLinkedListNode, FenwickTree, Graph, GraphEdge, GraphVertex, HashTable, Heap, InsertionSort, KMeans, Knapsack, KnapsackItem, LinkedList, LinkedListNode, MaxHeap, MergeSort, MinHeap, PixelColor, PixelCoordinate, PolynomialHash, PriorityQueue, Queue, QuickSort, QuickSortInPlace, RadixSort, RedBlackTree, SegmentTree, SelectionSort, ShellSort, SimplePolynomialHash, Sort, Stack, Trie, TrieNode, accumulatorBestTimeToBuySellStocks, articulationPoints, backtrackingJumpGame, bellmanFord, bfMaximumSubarray, bfRainTerraces, bfTravellingSalesman, binarySearch, bitLength, bitsDiff, bitsToFloat16, bitsToFloat32, bitsToFloat64, breadthFirstSearch, breadthFirstSearch1, btPowerSet, btUniquePaths, bwPowerSet, caesarCipherDecrypt, caesarCipherEncrypt, cartesianProduct, classicPolynome, clearBit, combinationSum, combineWithRepetitions, combineWithoutRepetitions, countSetBits, dcMaximumSubarraySum, decodeRailFenceCipher, degreeToRadian, depthFirstSearch, depthFirstSearch1, depthFirstSearchRecursive, detectDirectedCycle, detectUndirectedCycle, detectUndirectedCycleUsingDisjointSet, dijkstra, discreteFourierTransform, divideByTwo, dpBestTimeToBuySellStocks, dpBottomUpJumpGame, dpLongestIncreasingSubsequence, dpMaximumSubarray, dpRainTerraces, dpTopDownJumpGame, dpUniquePaths, dqBestTimeToBuySellStocks, encodeRailFenceCipher, euclideanAlgorithm, euclideanAlgorithmIterative, euclideanDistance, eulerianPath, factorial, factorialRecursive, fastFourierTransform, fastPowering, fibonacci, fibonacciNth, fibonacciNthClosedForm, fisherYates, floatAs32BinaryString, floatAs64BinaryString, fullAdder, getBit, getPixel, graphBridges, greedyJumpGame, hammingDistance, hanoiTower, hardyRamanujan, hillCipherDecrypt, hillCipherEncrypt, hornerMethod, integerPartition, interpolationSearch, inverseDiscreteFourierTransform, isEven, isPalindrome, isPositive, isPowerOfTwo, isPowerOfTwo1, isPowerOfTwoBitwise, jumpSearch, kNN, knightTour, knuthMorrisPratt, kruskal, leastCommonMultiple, linearSearch, liuHui, longestCommonSubsequence, longestCommonSubstring, Matrix$1 as matrix, multiply, multiplyByTwo, multiplyUnsigned, nQueens, pascalTriangle, pascalTriangleRecursive, peakvalleyBestTimeToBuySellStocks, permutateWithRepetitions, permutateWithoutRepetitions, prim, primeFactors, rabinKarp, radianToDegree, recursiveStaircaseBF, recursiveStaircaseDP, recursiveStaircaseIT, recursiveStaircaseMEM, regularExpressionMatching, resizeImageWidth, reverseTraversal, setBit, setPixel, shortestCommonSupersequence, sieveOfEratosthenes, squareMatrixRotation, squareRoot, stronglyConnectedComponents, switchSign, topologicalSort, traversal, trialDivision, uniquePaths, weightedRandom, zAlgorithm };
export as namespace feng3d;
