import { TrieNode } from './TrieNode';

// Character that we will use for trie tree root.
const HEAD_CHARACTER = '*';

/**
 * 字典树
 */
export class Trie
{
  head: TrieNode<string>;

  constructor()
  {
    this.head = new TrieNode(HEAD_CHARACTER);
  }

  /**
   * 添加单词
   *
   * @param word 单词
   * @return 自身
   */
  addWord(word: string)
  {
    const characters = Array.from(word);
    let currentNode = this.head;

    for (let charIndex = 0; charIndex < characters.length; charIndex += 1)
    {
      const isComplete = charIndex === characters.length - 1;
      currentNode = currentNode.addChild(characters[charIndex], isComplete);
    }

    return this;
  }

  /**
   * 删除单词
   *
   * @param word 单词
   * @return 自身
   */
  deleteWord(word: string)
  {
    const depthFirstDelete = (currentNode: TrieNode<string>, charIndex = 0) =>
    {
      if (charIndex >= word.length)
      {
        // Return if we're trying to delete the character that is out of word's scope.
        return;
      }

      const character = word[charIndex];
      const nextNode = currentNode.getChild(character);

      if (!nextNode)
      {
        // Return if we're trying to delete a word that has not been added to the Trie.
        return;
      }

      // Go deeper.
      depthFirstDelete(nextNode, charIndex + 1);

      // Since we're going to delete a word let's un-mark its last character isCompleteWord flag.
      if (charIndex === (word.length - 1))
      {
        nextNode.isCompleteWord = false;
      }

      // childNode is deleted only if:
      // - childNode has NO children
      // - childNode.isCompleteWord === false
      currentNode.removeChild(character);
    };

    // Start depth-first deletion from the head node.
    depthFirstDelete(this.head);

    return this;
  }

  /**
   *
   * @param word 单词
   */
  suggestNextCharacters(word: string)
  {
    const lastCharacter = this.getLastCharacterNode(word);

    if (!lastCharacter)
    {
      return null;
    }

    return lastCharacter.suggestChildren();
  }

  /**
   * Check if complete word exists in Trie.
   *
   * @param word 单词
   */
  doesWordExist(word: string)
  {
    const lastCharacter = this.getLastCharacterNode(word);

    return !!lastCharacter && lastCharacter.isCompleteWord;
  }

  /**
   *
   * @param word 单词
   */
  getLastCharacterNode(word: string)
  {
    const characters = Array.from(word);
    let currentNode = this.head;

    for (let charIndex = 0; charIndex < characters.length; charIndex += 1)
    {
      if (!currentNode.hasChild(characters[charIndex]))
      {
        return null;
      }

      currentNode = currentNode.getChild(characters[charIndex]);
    }

    return currentNode;
  }
}
